
/*
 * $Id: ObjLifeCycle.cc,v 1.6 1999/05/26 19:03:18 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include "Clock.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "ObjTimes.h"
#include "ObjLifeCycle.h"
#include "gadgets.h"


ObjLifeCycle::ObjLifeCycle():
	theBday(0),
	theModTime(0), theNowExpTime(0), theModExpTime(0),
	theModVar(0), theShowModRatio(0),
	theNowExpRatio(0), theModExpRatio(0), theShowExpRatio(0) {
}

ObjLifeCycle::~ObjLifeCycle() {
	delete theModTime;
	delete theNowExpTime;
	delete theModExpTime;
}

void ObjLifeCycle::bday(RndDistr *aDistr) {
	delete theBday;
	theBday = aDistr;
	if (theBday)
		theBday->rndGen(&theRng);
}

void ObjLifeCycle::modVar(double aVar) {
	theModVar = aVar;
}

void ObjLifeCycle::modTime(RndDistr *aDistr, double aRatio) {
	Assert(aDistr || aRatio == 0);
	delete theModTime;
	theModTime = aDistr;
	if (theModTime) {
		Assert(theBday);
		theModTime->rndGen(&theRng);
	}
	theShowModRatio = aRatio;
}

void ObjLifeCycle::nowExpTime(RndDistr *aDistr, double aRatio) {
	Assert(aDistr || aRatio == 0);
	delete theNowExpTime;
	theNowExpTime = aDistr;
	theNowExpRatio = aRatio;
	theShowExpRatio = theNowExpRatio + theModExpRatio;
	if (theNowExpTime)
		theNowExpTime->rndGen(&theRng);
}

void ObjLifeCycle::modExpTime(RndDistr *aDistr, double aRatio) {
	Assert(aDistr || aRatio == 0);
	delete theModExpTime;
	theModExpTime = aDistr;
	theModExpRatio = aRatio;
	theShowExpRatio = theNowExpRatio + theModExpRatio;
	if (theModExpTime)
		theModExpTime->rndGen(&theRng);
}

// calculates modification and expiration times for a given object
// note: result times must be rounded to second resolution or
//       comparison of HTTP-equal dates may not work!
void ObjLifeCycle::calcTimes(int oid, ObjTimes &times) {
	// must have a mod time model to proceed
	if (!theModTime) {
		times.reset();
		return;
	}

	// seed random generator with object id
	// note: both distributions is linked to this rng
	theRng.seed(RndPermut(oid));

	// all time computations are in [sec]

	// used a lot
	const int now = TheClock.time().sec();

	// several "persistent" choices
	const double expModelSeed = theRng();
	times.doShowLmt = theRng() < theShowModRatio;
	int bday = (int)theBday->trial(); // absolute or relative
	const int cycle = (int)theModTime->trial();

	// also persistent: expiration gap (used below)
	// the delta is persistent because that's how real servers
	// have it configured: mod + const_delta OR now + const_delta
	const bool knownExp = expModelSeed < theShowExpRatio;
	const int expGap = !knownExp ? 0 :
		expModelSeed < theNowExpRatio ?
			(int)theNowExpTime->trial() :
			(int)(cycle*theModExpTime->trial()); // convert #cycles to sec

	// adjust birthday: (relative "|n| secs since start" to absolute)
	if (bday <= 0)
		bday += Clock::TheStartTime.sec();

	// find lmt (without random, cycle dependent delta)
	const int cycleCnt = now > bday ? (now - bday)/cycle : 0;

	// compute lmt, adding random, cycle dependent delta (changes seed!)
	int lmt = bday + lmtDelta(oid, cycleCnt, cycle);
	if (lmt > now) // oops, random overshoot!
		lmt = bday + lmtDelta(oid, cycleCnt-1, cycle); // previous cycle

	// be sure lmt does not precede bday
	times.theLmt = Time::Sec(Max(bday, lmt));

	// calculate expiration time if needed
	if (knownExp) {
		if (expModelSeed < theNowExpRatio)
			times.theExp = TheClock + Time::Sec(expGap);
		else
			times.theExp = times.theLmt + Time::Sec(expGap);
	} else
		times.theExp = Time();
}

// compute lmt delta for a given cycle (with a random component)
// note: cycleCnt == number of full (completed) cycles
int ObjLifeCycle::lmtDelta(int oid, int cycleCnt, int cycleLen) {
	if (cycleCnt <= 0)
		return 0;
	theRng.seed(RndPermut(oid, cycleCnt));
	const double drift = cycleLen/2.0;
	return cycleCnt*cycleLen + (int) (theModVar * theRng(-drift, +drift));
}
