
/*
 * $Id: StatPing.cc,v 1.5.2.7 1999/06/14 21:07:59 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <iomanip.h>

#include "Clock.h"
#include "Socket.h"
#include "StatPing.h"
#include "StatPhaseMgr.h"
#include "Progress.h"
#include "ILog.h"
#include "LogComment.h"
#include "logTags.h"
#include "NotifMsg.h"
#include "gadgets.h"
#include "globals.h"

/* HRStat */

HRStat::HRStat() {
	reset();
}

void HRStat::reset() {
	theHitCnt = theMissCnt = 0;
	theHitSz = theMissSz = BigSize();
}

OLog &HRStat::store(OLog &log) const {
	// XXX: this hack must go away with new log format!
	const Size hs = (int)(theHitSz.byted()/(1024*1024)); // convert to MB
	const Size ms = (int)(theMissSz.byted()/(1024*1024)); // convert to MB
	return log << theHitCnt << theMissCnt << hs << ms;
}

ILog &HRStat::load(ILog &log) {
	// XXX: this hack must go away with new log format!
	Size hs, ms;
	log >> theHitCnt >> theMissCnt >> hs >> ms;
	theHitSz = BigSize::MB(hs); // convert from MB
	theMissSz = BigSize::MB(ms); // convert from MB
	return log;
}

double HRStat::dhr() const {
	return Ratio(theHitCnt, theHitCnt + theMissCnt);
}

double HRStat::bhr() const {
	const BigSize denom = theHitSz + theMissSz;
	return denom ? theHitSz / denom : -1;
}

double HRStat::dhp() const {
	return Percent(theHitCnt, theHitCnt + theMissCnt);
}

double HRStat::bhp() const {
	const BigSize denom = theHitSz + theMissSz;
	return denom ? 100 * (theHitSz / denom) : -1;
}

HRStat &HRStat::operator +=(const HRStat &hr) {
	theHitCnt += hr.theHitCnt;
	theMissCnt += hr.theMissCnt;
	theHitSz += hr.theHitSz;
	theMissSz += hr.theMissSz;
	return *this;
}


/* StatPingRec */

StatPingRec::StatPingRec() {
	theSelCnt = theRdCnt = theWrCnt = 0;
	theRdSz = theWrSz = 0;
	theErrCnt = 0;
}

OLog &StatPingRec::store(OLog &log) const {
	Assert(theDuration >= 0);
	return log
		<< theXactLvl << theOpenLvl << theEstbLvl
		<< theSelCnt << theRdCnt << theWrCnt
		<< theRdSz << theWrSz
		<< theRealHR
		<< theErrCnt
		<< theDuration
		// XXX: << theIdealHR
		;
}

ILog &StatPingRec::load(ILog &log) {
	return log
		>> theXactLvl >> theOpenLvl >> theEstbLvl
		>> theSelCnt >> theRdCnt >> theWrCnt
		>> theRdSz >> theWrSz
		>> theRealHR
		>> theErrCnt
		>> theDuration
		// XXX: >> theIdealHR
		;
}

void StatPingRec::restart() {
	theXactLvl.restart();
	theOpenLvl.restart();
	theEstbLvl.restart();
	theSelCnt = theRdCnt = theWrCnt = 0;
	theRdSz = theWrSz = 0;
	theRealHR.reset();
	theIdealHR.reset();
	theChbR.reset();
	theErrCnt = 0;
}

// XXX: we should have Rate type, not double
double StatPingRec::reqRate() const {
	return theDuration > 0 ?
		Ratio(theXactLvl.incCnt(), theDuration.secd()) : -1;
}

double StatPingRec::repRate() const {
	return theDuration > 0 ?
		Ratio(theXactLvl.decCnt(), theDuration.secd()) : -1;
}

StatPingRec &StatPingRec::operator +=(const StatPingRec &s) {
	theXactLvl.mergeWith(s.theXactLvl);
	theOpenLvl.mergeWith(s.theOpenLvl);
	theEstbLvl.mergeWith(s.theEstbLvl);

	theSelCnt += s.theSelCnt;
	theRdCnt += s.theRdCnt;
	theWrCnt += s.theWrCnt;
	theRdSz += s.theRdSz;
	theWrSz += s.theWrSz;

	theRealHR += s.theRealHR;
	theIdealHR += s.theIdealHR;
	theChbR += s.theChbR;

	theErrCnt += s.theErrCnt;

	// do not update duration

	return *this;
}

/* StatPing */

StatPing::StatPing(Time aPeriod): StatClockInt(aPeriod) {
}

double TheOptReqRate = -1; // XXX
NetAddr TheMonitor; // XXX
String TheRunLabel; // XXX

void StatPing::flush() {
	theDuration = TheClock - theIntStart;

	OLog &log = TheLog();
	log << bege(lgStatPing);
	StatPingRec::store(log);
	theIdealHR.store(log); // XXX
	theChbR.store(log); // XXX
	log << ende;

	if (theDuration > 0)
		report();

	// XXX: wrong location and requires use of extern TheOptReqRate, fix!
	if (TheOptReqRate > 0) {
		static int TheComplCnt = 0;
		static int TheSubmCnt = 0;
		static int TheSuccessCnt = 0;
		static bool InRecoverMode = false;
		static Time TheLastCheckTime, TheNextCheckTime;
		TheComplCnt += theXactLvl.decCnt();
		TheSubmCnt += theXactLvl.incCnt();
		if (TheClock.time() >= TheNextCheckTime) {

			const double sec = (TheClock - TheLastCheckTime).secd();
			if (TheLastCheckTime >= 0) {
				if (abs(TheSubmCnt - TheComplCnt) < 0.01 * Min(TheSubmCnt,TheComplCnt)) {
					TheSuccessCnt = InRecoverMode ? 1 : Max(TheSuccessCnt+1, 1);
					TheOptReqRate += TheSuccessCnt*0.01*TheOptReqRate;
					InRecoverMode = false;
				} else
				if (TheComplCnt < TheSubmCnt) {
					TheSuccessCnt = Max(TheSuccessCnt, 1);
					TheOptReqRate -= TheSuccessCnt*0.01*TheOptReqRate;
					TheSuccessCnt--;
					const double max_rate = Ratio(TheSubmCnt + TheComplCnt, sec)/2;
					if (TheOptReqRate > max_rate)
						TheOptReqRate = max_rate;
					InRecoverMode = true;
				}
			}

			Comment(4) << "avg rates: req: " << Ratio(TheSubmCnt, sec) << " rep: " << Ratio(TheComplCnt, sec) << "; target: " << TheOptReqRate << endc;
			TheNextCheckTime = TheClock + Time::Sec(10);
			TheLastCheckTime = TheClock;
			TheComplCnt = TheSubmCnt = 0;
			//TheOptReqRate = -1;
		}
	}

	// XXX: wrong location
	if (TheMonitor) {
		static Time TheLastCheckTime, TheNextCheckTime;
		static Socket TheMon;

		if (TheClock.time() >= TheNextCheckTime) {

			// first time initializations
			if (TheNextCheckTime < 0) {
				bool ok = true;

				ok = ok && TheMon.create(AF_INET, SOCK_DGRAM, 0);
				ok = ok && TheMon.connect(TheMonitor);
				ok = ok && TheMon.blocking(false);

				if (!ok) {
					TheMon.close();
					Comment(4) << "warning: disabling notification; possible cause: " << error(errno) << endc;
				}
			}

			if (TheMon) {
				StatusNotifMsg msg(TheRunLabel);

				msg.theStartTime = TheProgress.startTime();
				msg.theSndTime = TheClock;
				msg.theRespTime = Time::Msec((int)TheSoOftenPinger.xactLifeTime().stats().mean());
				msg.theTput = Ratio(theXactLvl.decCnt(), theDuration.secd());
				msg.theBwidth = Ratio(theRdSz + theWrSz, theDuration.secd());
				msg.theDHR = theRealHR.dhr();
				msg.theXactCnt = TheProgress.theXactComplCnt;
				msg.theSockCnt = Socket::Level();

				// write, ignoring errors
				msg.hton();
				TheMon.write(&msg, sizeof(msg));
			}

			TheLastCheckTime = TheClock;
			TheNextCheckTime = TheClock + Time::Sec(20);
		}
	}
}

void StatPing::restart() {
	StatClockInt::restart();
	StatPingRec::restart();
}

void StatPing::report() const {
	Comment(2)
		<< "i-" << TheStatPhaseMgr->name()
		<< ' ' << setw(6) << TheProgress.theXactComplCnt
		<< ' ' << setw(6) << Ratio(theXactLvl.decCnt(), theDuration.secd())
		<< ' ' << setw(6) << theRealHR.dhp()
		<< ' ' << setw(3) << theErrCnt
		<< ' ' << setw(3) << Socket::Level()
		<< endc;
}
