
/*
 * $Id: World.cc,v 1.2 1999/05/26 19:03:21 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <math.h>
#include <fstream.h>

#include "PopModel.h"
#include "ChbModel.h"
#include "TmlModel.h"
#include "HitModel.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "World.h"
#include "gadgets.h"

World TheWorld;

World::World(): thePopMdl(0), theChbMdl(0), theTmlMdl(0), theHitMdl(0),
	theLastChbId(0), theLastUncId(0) {
	// thePopMdl must be set by the client (there may be several models)
	theChbMdl = new ChbModel;
	theTmlMdl = new TmlModel;
	theHitMdl = new HitModel;
}

World::~World() {
	while (theUrls.count())
		delete[] theUrls.pop();
	delete thePopMdl;
	delete theChbMdl;
	delete theTmlMdl;
	delete theHitMdl;
}


int World::genOid(OidRec &rec) {
	Assert(thePopMdl);

	replanOid = true;

	// hit or miss?
	rec.ohit = theLastChbId && theHitMdl->hitIt();

	int oid = 0;
	if (rec.ohit) { // hit
		// repeat some cachable Oid
		oid = repeatOid(theLastChbId, true);
		rec.cachable = true;
	} else { // miss
		// cachable?
		rec.cachable = theChbMdl->cacheIt();

		if (rec.cachable) {
			// cachable miss; must emit new cachable Oid
			oid = Id2Oid(++theLastChbId, true);
		} else {
			// XXX: better have different Hit Models for un/cachable
			// XXX: better generate a random variate once and reuse it
			const bool repeat = theLastUncId && theHitMdl->hitIt();
			if (repeat) {
				// repeat some uncachable Oid
				oid = repeatOid(theLastUncId, false);
			} else {
				// emit new uncachable Oid
				oid = Id2Oid(++theLastUncId, false);
			}
		}
	}

	// update history
	if (theTmlMdl->enabled()) {
		if (replanOid)
			theTmlMdl->addOid(oid);
		theTmlMdl->advance();
	}

	Assert(oid > 0);
	rec.oid = oid; // XXX: either store it in rec or return it
	return oid;
}

// generate an old Oid using tmp locality model (if any)
int World::repeatOid(int lastId, bool cachable) {
	// XXX: better have different Tml Models for un/cachable
	if (theTmlMdl->enabled()) {
		if (const int oid = theTmlMdl->pickOid(cachable))
			return oid;
	}

	replanOid = false;
	return Id2Oid(thePopMdl->repeatId(lastId), cachable);
}

void World::prefill(int count) {
	Assert(count > 0);
	Assert(!theLastChbId); // we have not started generating Ids yet
	theLastChbId = count;

	// let the popularity and tmp locality models to prefill the plan
	if (theTmlMdl->enabled()) {

		// when to stop?!
		// a mean plus three std_dev
		int depth = (int) (theTmlMdl->distr()->mean() + 3*theTmlMdl->distr()->sdev());
		// adjust for hit rate
		depth += (int) (depth * (1 - theHitMdl->hr())); 
		
		Comment(3) << "warming up the tmp loc plan; oids: " << count << " slots: " << depth << endc;

		while (depth-- > 0) {
			// XXX: this emulates genOid except we do not add misses
			if (theHitMdl->hitIt())
				theTmlMdl->addOid(repeatOid(theLastChbId, true));
			theTmlMdl->advance();
		}

		Comment(3) << "warmed up the tmp loc plan; current depth: " << theTmlMdl->curPlanDepth() << endc;
	}
}

bool World::loadUrls(const char *fname) {
	char url[8*1024];

	Assert(fname);

	ifstream is(fname);
	Size sz = 0;
	int lineCount = 0;
	int urlCount = 0;

	Comment(1) << "loading urls from '" << fname << "'" << endc;

	while (is && is.getline(url, sizeof(url))) {
		lineCount++;
		const int len = strlen(url);
		if (len && *url == '/') {
			theUrls.append(StrDup(url));
			sz += sizeof(char*) + len + 1;
			urlCount++;
		} else
		if (len && *url != '#') {
			Comment << fname << ':' << lineCount << " unknown url format near " << url << endc;
			return false;
		}
	}

	is.close();

	Comment(1)
		<< "loaded " << urlCount 
		<< " urls from '" << fname << "'"
		<< " (" << lineCount << " lines)"
		<< " consumed ~ " << sz/(1024*1024.) << " MB" << endc;

	return urlCount > 0;
}

bool World::genObjIds(int count, int order, bool unique) {
	Assert(0); // XXX: broken? by latest additions?
	// XXX: fix "int oid(int offset, bool &repeat) const"!
	OidRec rec;

	// pump up so we do not get wrapped at the end of the experiment
	// the latter may cause hits with miss-only workload
	count += count/20; // 5%

	theObjIds.stretch(count);
	for (int i = 0; i < count; ++i)
		theObjIds.append(genOid(rec));

	// remember positions if needed
	if (unique) {
		theUnqIds.stretch(count);
		for (int i = 1; i <= count; ++i)
			theUnqIds.append(i);
	}

	// perform permutation if needed
	if (order != 1) { // note: order == 1 -- initial order
		// decide if permutation of order is needed
		RndGen rng(-1/*order*/);
		for (int i = 0; i < count; ++i) {
			const int newOff = rng(0, count);
			theObjIds.swap(i, newOff);
			if (unique)
				theUnqIds.swap(i, newOff);
		}
	}

	return theObjIds.count() >= count;
}

ostream &World::report(ostream &os) const {
	return os << "Unique oids:"
		<< " cachable: " << theLastChbId+1
		<< " uncachable: " << theLastUncId+1
		;
}
