
/*
 * $Id: httpHdrs.cc,v 1.2 1999/04/30 04:47:14 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include "String.h"
#include "StrIdentifier.h"
#include "HttpDate.h"
#include "httpHdrs.h"
#include "gadgets.h"


/* internal type to store static parsing info */
class MsgHdrParsTab {
	public:
		MsgHdrParsTab();
		~MsgHdrParsTab();

	public:
		StrIdentifier *ids;
		Array<MsgHdr::Parser> *parsers;
};


MsgHdrParsTab *ReqHdr::TheParsTab = 0;
MsgHdrParsTab *RepHdr::TheParsTab = 0;


/* MsgHdr */

MsgHdr::MsgHdr(const MsgHdrParsTab &aTab): theParsTab(aTab) {
	reset();
}

void MsgHdr::reset() {
	theHdrSize = 0;
	theContSize = -1;
	theXactId.clear();
	theHttpVer = -1;
	theKeepAliveConn = kaDefault;

	theBufBeg = theBufEnd = theSrchPtr = 0;
	theFields.reset();
	theSrchState = ssFirst;

	// do not reset parsing tables
}

bool MsgHdr::keepAlive() const {
	if (theHttpVer >= 1.1) // 1.1: keep unless told otherwise
		return theKeepAliveConn != kaNo;
	else                   // 1.0: keep if explicitly told so
		return theKeepAliveConn == kaYes;
}

// note: buf does not have to be zero-terminated!
bool MsgHdr::parse(const char *buf, Size sz) {
	if (!theBufBeg) { // have not started the search yet
		theBufBeg = theSrchPtr = buf;
		Assert(theSrchState == ssFirst);
	} else {           // continue search
		Assert(theBufBeg == buf);
		Assert(theSrchState != ssFound);
	}

	theBufEnd = buf + sz; // to be refined later
	do {
		// search for LF
		if (theSrchState == ssFirst) {
			do {
				if (*theSrchPtr++ == '\n') {
					theSrchState = ssSkip;
					break;
				}
			} while (theSrchPtr < theBufEnd);
		}

		// LF after skipping optional CRs means and-of-headers
		while (theSrchState == ssSkip && theSrchPtr < theBufEnd) {
			if (*theSrchPtr == '\n') {
				theSrchState = ssFound;
			} else
			if (*theSrchPtr != '\r') {
				theFields.append(theSrchPtr); // start of a header!
				theSrchState = ssFirst;
			}
			++theSrchPtr;
		}
	} while (theSrchPtr < theBufEnd && theSrchState != ssFound);

	if (theSrchState != ssFound)
		return false;

	// found end-of-headers!
	theBufEnd = theSrchPtr;
	theHdrSize = theBufEnd - theBufBeg;

	// now parse known fields
	// luckily, we already know field starts!
	parseFields();
	return true;
}

void MsgHdr::parseFields() {
	const char *eoh = theBufEnd;
	parseRLine(theBufBeg, theFields.count() ? theFields[0] : theBufEnd);

	for (int i = theFields.count()-1; i >= 0; --i) {
		const char *hdr = theFields[i];
		const int len = eoh-hdr; // approximate (includes crlfs)
		const int id = theParsTab.ids->lookup(hdr, len);
		if (id > 0) {
			const char *val = hdr + theParsTab.ids->string(id).len();
			while (isspace(*val)) ++val;
			Parser p = theParsTab.parsers->item(id);
			(this->*p)(val, eoh);
		}
		eoh = hdr;
	}
}

bool MsgHdr::parseContLen(const char *buf, const char *) {
	theContSize = xatoi(buf, -1);
	return theContSize >= 0;
}

bool MsgHdr::parseXXact(const char *buf, const char *) {
	if (const char *b = strchr(buf, '.'))
		if (const char *c = strchr(++b, ':'))
				return theXactId = UniqId(xatoi(buf), xatoi(b), xatoi(c+1));
	return false;
}

bool MsgHdr::parseConnection(const char *buf, const char *) {
	// XXX: Connection may have a _list_ of options
	if (!strncasecmp("close", buf, 5))
		theKeepAliveConn = kaNo;
	else
	if (!strncasecmp("keep", buf, 4))
		theKeepAliveConn = kaYes;
	else
		return false;
	return true;
}

// adds definitions common to replies and requests
void MsgHdr::Configure(MsgHdrParsTab &tab) {
	AddParser("Content-Length:", &MsgHdr::parseContLen, tab);
	AddParser("X-Xact:", &MsgHdr::parseXXact, tab);
	AddParser("Connection:", &MsgHdr::parseConnection, tab);
	AddParser("Proxy-Connection:", &MsgHdr::parseConnection, tab);
}

int MsgHdr::AddParser(const String &field, Parser parser, MsgHdrParsTab &where) {
	const int id = where.ids->add(field);
	where.parsers->put(parser, id);
	return id;
}

// these should never be called
inline bool dontCallMe() { Assert(0); return false; }
bool MsgHdr::parseXEtag(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parsePragma(const char *, const char *)  { return dontCallMe(); }
bool MsgHdr::parseCControl(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseIMS(const char *, const char *) { return dontCallMe(); }


/* ReqHdr */

ReqHdr::ReqHdr(): MsgHdr(*TheParsTab), theObjId(-1), theETag(-1) {
}

void ReqHdr::reset() {
	MsgHdr::reset();
	theIMS = Time();
	theObjId = theETag = -1;
}

bool ReqHdr::parseRLine(const char *buf, const char *) {
	if (!strncasecmp("GET ", buf, 4)) {
		buf += 4;
		if (const char *h = strchr(buf, '_')) // XXX: fix all for !0-term!!
			theObjId = xatoi(buf = h + 1, -1);
		if (const char *h = strchr(buf, ' '))
			isDbl(h + 6, theHttpVer);
		return true;
	}
	return false;
}

bool ReqHdr::parseXEtag(const char *buf, const char *) {
	if (const char *h = strchr(buf, '"'))
		theETag = xatoi(h+1, -1);
	else
		return false;
	return true;
}

bool ReqHdr::parseIMS(const char *buf, const char *eoh) {
	theIMS = HttpDateParse(buf, eoh - buf);
	return theIMS >= 0;
}

void ReqHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser("X-Etag:", &MsgHdr::parseXEtag, *TheParsTab);
	AddParser("If-Modified-Since:", &MsgHdr::parseIMS, *TheParsTab);
}


/* RepHdr */

RepHdr::RepHdr(): MsgHdr(*TheParsTab), isCachable(true) {
}

void RepHdr::reset() {
	MsgHdr::reset();
	isCachable = true;
}

bool RepHdr::parseRLine(const char *buf, const char *) {
	// what is the actual version format in the specs??
	if (!strncasecmp("HTTP/", buf, 5))
		return isDbl(buf+5, theHttpVer);
	else
		return false;
}

bool RepHdr::parsePragma(const char *buf, const char *) {
	// XXX: Pragma may have a _list_ of options
	if (!strncasecmp("no-cache", buf, 8))
		isCachable = false;
	else
		return false;
	return true;
}

bool RepHdr::parseCControl(const char *buf, const char *) {
	// XXX: Cache-Control may have a _list_ of options
	if (!strncasecmp("no-cache", buf, 8))
		isCachable = false;
	else
		return false;
	return true;
}

void RepHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser("Pragma:", &MsgHdr::parsePragma, *TheParsTab);
	AddParser("Cache-Control:", &MsgHdr::parseCControl, *TheParsTab);
}


/* MsgHdrParsTab */

MsgHdrParsTab::MsgHdrParsTab() {
	ids = new StrIdentifier;
	parsers = new Array<MsgHdr::Parser>;
}

MsgHdrParsTab::~MsgHdrParsTab() {
	delete ids; ids = 0;
	delete parsers; parsers = 0;
}



/* initialization */

int HttpHdrsInit::theUseCount = 0;

void HttpHdrsInit::init() {
	ReqHdr::Configure();
	RepHdr::Configure();
}

void HttpHdrsInit::clean() {
	Assert(theUseCount == 0);
	// XXX: protected status prevents destruction
	//ReqHdr::TheParsTab.destroy();
	//RepHdr::TheParsTab.destroy();
}
