
/*
 * $Id: polyclt.cc,v 1.7.2.3 1999/06/14 21:08:00 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <new.h>

// IRIX demands _BSD_SIGNALS to be defined for signal() 
// catching function declaration to have "int" as the first argument
#ifndef _BSD_SIGNALS
#define _BSD_SIGNALS
#endif
#include <signal.h>


#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "Clock.h"
#include "FileScanner.h"
#include "AsyncClt.h"
#include "SyncClt.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "StatPhaseMgr.h"
#include "EphPortMgr.h"
#include "ExpPortMgr.h"
#include "OLog.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "Socket.h"
#include "CltOpts.h"
#include "CltConnMgr.h"
#include "PopModel.h"
#include "ChbModel.h"
#include "TmlModel.h"
#include "HitModel.h"
#include "World.h"

#include "gadgets.h"
#include "rndDistrs.h"
#include "logTags.h"
#include "polyErrors.h"
#include "shutdown.h"
#include "globals.h"

// runs one select loop cycle
static
void step() {
	Time timeout = TheAlarmClock.timeLeft();
	//cerr << here << " select tout: " << timeout << " AlarmClockClear: " << TheAlarmClock.clear() << endl;

	const int readyCount = TheFileScanner->scan(TheAlarmClock.clear() ? 0 : &timeout);
	CallStat(ThePinger, selected());
	//cerr << here << " select ready: " << readyCount << endl;

	if (readyCount < 0 && errno != EINTR)
		FatalError(errno);
}

static
void configureLogs(int prec) {
	if (TheOpts.theConsFileName && strcmp(TheOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theConsFileName);

	// set log file and log size
	Size logSz = TheOpts.theLogSize;
	if (TheOpts.theLogFileName) {
		TheLog().stream(TheOpts.theLogFileName, new ofstream(TheOpts.theLogFileName));
		// set default log size if needed
		if (logSz < 0)
			logSz = Size::MB(10);
	} else {
		// virtually disable logging (will use ctor default)
		if (logSz < 0)
			logSz = Size::MB(0);
	}
	Assert(logSz >= 0);
	TheLog().resize(Max(TheLog().capacity(), logSz)); // we cannot shrink log size
	TheOpts.theLogSize.set(TheLog().capacity());

	Comment.TheEchoLevel = TheOpts.theVerbLevel;

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

// set some general stuff and
// propogate cmd line options to corresponding objects
static
void configure() {

	configureLogs(2);

	atexit(&Shutdown);

	ErrorMgr::Configure(new PolyErrMgr);

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	// quit nicely on some signals
	signal(SIGINT, &ShutdownSignal);

	// set resource limits
	const int rlimit = SetMaxFD(GetMaxFD());
	if (rlimit > FD_SETSIZE)
		cerr << here << "warning: getrlimit(2) system call reports " << rlimit << " file descriptors while FD_SETSIZE #defines only " << FD_SETSIZE << "; using lower value." << endl;
	const int fdLimit = Min(FD_SETSIZE, rlimit);
	if (fdLimit <= 0)
		FatalError(errZeroFDLimit);
	TheFileScanner = TheOpts.theFileScanner.val();
	TheFileScanner->configure(fdLimit);

	Socket::TheMaxLevel = Max(0, fdLimit - 10 - fdLimit/33);
	if (TheOpts.theFDLimit >= 0 && TheOpts.theFDLimit < Socket::TheMaxLevel)
		Socket::TheMaxLevel = TheOpts.theFDLimit;
	TheOpts.theFDLimit.set(Socket::TheMaxLevel);

	Socket::IsEnabled = TheCltOpts.doIOs;

	// set random seeds
	RndPermutSeed(TheOpts.theRngSeed);
	RndGen::TheDefSeed = RndPermut(TheOpts.theRngSeed);

	TheStatPhaseMgr.goal(TheOpts.theGoal);

	// configure the World
	TheWorld.type(TheCltOpts.theWorld.theType);
	TheWorld.id(TheCltOpts.theWorld.theId);
	TheWorld.theChbMdl->ratios(TheCltOpts.theWorld.theChbRatio, TheCltOpts.theWorld.theHitRatio);
	TheWorld.theHitMdl->hr(TheCltOpts.theWorld.theHitRatio);
	TheWorld.thePopMdl = TheCltOpts.theWorld.thePopMdl.model();
	if (TheCltOpts.theWorld.theTmlDistr) {
		TheWorld.theTmlMdl->distDistr(TheCltOpts.theWorld.theTmlDistr.condDistr());
		TheWorld.theTmlMdl->searchDelta(TheCltOpts.theWorld.theTmlDelta);
		TheWorld.theTmlMdl->planDepth(TheCltOpts.theWorld.theTmlDepth);
	}
	// note: prefilling, if any, is done after cfg and report

	if (TheCltOpts.theWorld.theUrlsFileName &&
		!TheWorld.loadUrls(TheCltOpts.theWorld.theUrlsFileName)) {
		Comment << "failed to load urls from `" << TheCltOpts.theWorld.theUrlsFileName << "'" << endc;
		exit(-2);
	}
	if (TheCltOpts.theWorld.theReqOrder >= 0) {
		Assert(TheOpts.theGoal.theXactComplCnt >= 0);
		TheWorld.genObjIds(TheOpts.theGoal.theXactComplCnt, TheCltOpts.theWorld.theReqOrder, TheCltOpts.theWorld.theUniqStart >= 0);
	}

	// explicit or ephemeral binding
	ThePortMgr = TheCltOpts.thePorts.set() ?
		new ExpPortMgr(TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		new EphPortMgr();

	if (TheCltOpts.theCoolPhase > 0)
		CoolOffTime(TheCltOpts.theCoolPhase);
}

int main(int argc, char *argv[]) {

	CmdLine cmd(Array<OptGrp*>() << &TheOpts << &TheCltOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate() || !TheCltOpts.validate())
		return -1;

	Clock::Update(false);

	configure();

	cmd.report();

	Comment(2) << "FDs: "
		<< GetCurFD() << " out of " << GetMaxFD()
		<< " FDs can be used; safeguard limit: " << Socket::TheMaxLevel 
		<< endc;

	const UniqId runId = UniqId::Create();
	Comment(1) << "run-id: " << runId << " pid: " << getpid() << endc;
	TheLog() << bege(lgRunId) << runId << ende;


	if (TheCltOpts.theWorld.thePrefilledCnt > 0)
		TheWorld.prefill(TheCltOpts.theWorld.thePrefilledCnt);

	CltConnMgr TheConnMgr;
	TheConnMgr.configure(
		TheCltOpts.thePxyAddr ? TheCltOpts.thePxyAddr : TheCltOpts.theOrgAddr, 
		TheCltOpts.theConnPoolCap,
		Max(TheCltOpts.theRobotCnt * (TheCltOpts.theReqRate ? 1024 : 1),
			(int)TheCltOpts.theConnPoolCap));

	Array<Client*> TheClients((int)TheCltOpts.theRobotCnt);

	RndDistr *thinkDistr = TheOpts.theXactThinkTime.condDistr();
	// request rate requires its own distribution
	if (TheCltOpts.theReqRate) {
		Assert(!thinkDistr); // cannot have both (checked in validate())
		// all clients should share one r.n.g.
		RndGen *rng = new RndGen;
		const double mean = TheCltOpts.theReqRate.intArrTime().secd();
		// all clients should share one distribution
		// XXX: we need distribution types other than exponential here
		thinkDistr = new ExpDistr(rng, mean);
		//thinkDistr = new ConstDistr(rng, mean);

		// XXX: kludge, fix me!
		if (TheCltOpts.maximizeReqRate && mean) {
			extern double TheOptReqRate;
			TheOptReqRate = 1/mean;
		}
	}

	// XXX: kludge, fix me!
	if (TheOpts.theMonitor) {
		extern NetAddr TheMonitor;
		extern String TheRunLabel;
		
		TheMonitor = TheOpts.theMonitor;
		TheRunLabel = TheOpts.theRunLabel;
	}

	if (TheCltOpts.theReqRate) {
		for (int c = 0; c < TheCltOpts.theRobotCnt; ++c) {
			AsyncClt *clt = new AsyncClt();
			clt->configure(&TheConnMgr, thinkDistr);
			TheClients.append(clt);
		}
	} else {
		for (int c = 0; c < TheCltOpts.theRobotCnt; ++c) {
			SyncClt *clt = new SyncClt();
			clt->configure(&TheConnMgr, thinkDistr);
			TheClients.append(clt);
		}
	}
	Assert(TheClients.count() == TheCltOpts.theRobotCnt);

	Clock::Update();
	Comment << "waiting for replies; goal: " << TheOpts.theGoal << endc;

	ThePinger.start();
	TheSoOftenPinger.start();

	TheConnMgr.start();

	for (int c = 0; c < TheClients.count(); ++c) {
		TheClients[c]->start();
	}

	while (!ShutdownNow()) {
		Clock::Update();
		ThePinger.checkpoint();
		TheSoOftenPinger.checkpoint();
		step();
		TheConnMgr.checkCcLevel();
	}

	return 0;
}
